/*
 * $Id: interrupt.h,v 1.1 1997/05/22 21:20:06 ettore Exp $
 *
 * This file is part of VICE, the Versatile Commodore Emulator.
 * See README for copyright notice.
 *
 * Written by
 *   Ettore Perazzoli (ettore@comm2000.it)
 *   Andre' Fachat (fachat@physik.tu-chemnitz.de)
 *
 * $Log: interrupt.h,v $
 * Revision 1.1  1997/05/22 21:20:06  ettore
 * Initial revision
 *
 */

#ifndef _INTERRUPT_H
#define _INTERRUPT_H

#include "types.h"

/* This handles the interrupt lines and the CPU alarms (i.e. events that happen
   at specified clock ticks during emulation). */

/* For maximum performance (these routines are small but used very often), we
   allow use of inlined functions.  This can be overridden by defining
   NO_INLINE (useful for debugging and profiling).  */
#if defined(NO_INLINE) && defined(INLINE_INTERRUPT_FUNCS)
#  define INLINE_INTERRUPT_FUNCS
#else
#  undef INLINE_INTERRUPT_FUNCS
#endif

/* Define the tick when the clock counter overflow must be prevented and what
   value should be subtracted from the counter when the tick is reached. */
#define PREVENT_CLK_OVERFLOW_TICK (~((CLOCK)0) - 0x100000)
#define PREVENT_CLK_OVERFLOW_SUB (PREVENT_CLK_OVERFLOW_TICK & ~((CLOCK)0xffff))

/* -------------------------------------------------------------------------- */

/* These are the available types of interrupt lines. */
enum cpu_int {
    IK_NONE = 1 << 0,
    IK_NMI = 1 << 1,
    IK_IRQ = 1 << 2,
    IK_RESET = 1 << 3,
    IK_TRAP = 1 << 4
};

/* This defines the type for a CPU alarm handler.  The passed argument is set
   to the number of ticks between the time of call and the time of
   scheduling. */
typedef int (cpu_alarm_handler_t)(long delay);

/* We not care about wasted space here, and make fixed-length large enough
   arrays since static allocation can be handled more easily... */
struct cpu_int_status {
    /* Number of possible alarms. */
    int num_alarms;

    /* Number of interrupt lines. */
    int num_ints;
    
    /* Next CPU alarm to be served.  If -1, means ``check for pending
       interrupts''. */
    int next_alarm;

    /* Clock tick when the next pending CPU alarm has to be served. */
    CLOCK next_alarm_clk;

    /* This is nonzero if we might have a pending interrupt. */
    int should_check_pending_interrupt;
    
    /* Value of clk when each CPU alarm has to be served. ~0 means `not
       active'. */
    CLOCK alarm_clk[0x100];
    
    /* Define, for each interrupt source, whether it has a pending interrupt
       (IK_IRQ, IK_NMI, IK_RESET and IK_TRAP) or not (IK_NONE). */
    enum cpu_int pending_int[0x100];

    /* CPU alarm handlers. */
    cpu_alarm_handler_t *alarm_handler[0x100];

    /* Number of active IRQ lines. */
    int nirq;

    /* Number of active NMI lines. */
    int nnmi;
    /* Previous number of active NMI lines.  This is used to detect the 0->1
       transition. */
    int last_nnmi;

    /* If 1, do a RESET. */
    int reset;

    /* If 1, call the trapping function. */
    int trap;

    /* Debugging function. */
    void (*trap_func)(ADDRESS);
    
    /* IRQ/NMI strobe. */
    int strobe_status;
};

/* -------------------------------------------------------------------------- */

/* If we do not want inline interrupts, these are compiled once when included
   in maincpu.c. */

#ifdef INLINE_INTERRUPT_FUNCS
#  define _INT_FUNC inline static
#else
#  define _INT_FUNC
#endif

#if defined(INLINE_INTERRUPT_FUNCS) || defined(_MAINCPU_C)

#include <string.h>		/* memset() */

/* Initialization. */
_INT_FUNC void cpu_int_status_init(struct cpu_int_status *cs, int num_ints, 
 				   int num_alarms)
{
    int i;
    
    memset(cs, 0, sizeof(struct cpu_int_status));
    cs->num_ints = num_ints;
    cs->num_alarms = num_alarms;
    for (i = 0; i < 0x100; i++)
	cs->alarm_clk[i] = ~((CLOCK)0);
}

/* Set the IRQ line state. */
_INT_FUNC void set_irq(struct cpu_int_status *cs, int int_num, int value)
{
    if (value) {		/* Trigger the IRQ. */
	if (!(cs->pending_int[int_num] & IK_IRQ)) {
	    cs->nirq++;
	    cs->should_check_pending_interrupt++;
	    cs->pending_int[int_num] |= IK_IRQ;
	}
    } else {			/* Remove the IRQ condition. */
	if (cs->pending_int[int_num] & IK_IRQ) {
	    if (cs->nirq) {
		cs->nirq--;
		cs->should_check_pending_interrupt--;
		cs->pending_int[int_num] &= ~IK_IRQ;
	    } else
		printf("set_irq(): wrong nirq!\n");
	}
    }
}

/* Set the NMI line state. */
_INT_FUNC void set_nmi(struct cpu_int_status *cs, int int_num, int value)
{
    if (value) {		/* Trigger the NMI. */
	if (!(cs->pending_int[int_num] & IK_NMI)) {
	    cs->nnmi++;
	    cs->should_check_pending_interrupt++;
	    cs->pending_int[int_num] |= IK_NMI;
	}
    } else {			/* Remove the NMI condition. */
	if (cs->pending_int[int_num] & IK_NMI) {
	    if (cs->nnmi) {
		cs->nnmi--;
		/* When going down, set last_nnmi. This works because only the
	           0->1 flank has to be detected, not the 1->0 one.  Otherwise,
	           when removing an NMI, but not calling cpu_do_int() before
	           the next NMI occurs, the transition is not detected because
	           cpu_do_int() had no chance setting last_nnmi and the NMI is
	           missed.  */
		cs->last_nnmi = cs->nnmi;
		cs->should_check_pending_interrupt--;
		cs->pending_int[int_num] &= ~IK_NMI;
	    } else
		printf("set_nmi(): wrong nnmi!\n");
	}
    }
}

/* Change the interrupt line state: this can be used to change both NMI and IRQ
   lines.  It is slower than set_nmi() and set_irq(), but is left for backward
   compatibility (it works like the old setirq()). */
_INT_FUNC void set_int(struct cpu_int_status *cs, int int_num,
		       enum cpu_int value)
{
    set_nmi(cs, int_num, value & IK_NMI);
    set_irq(cs, int_num, value & IK_IRQ);
}

/* Strobe IRQ. */
_INT_FUNC void strobe_irq(struct cpu_int_status *cs)
{
    if (!(cs->strobe_status & IK_IRQ)) {
	cs->should_check_pending_interrupt++;
	cs->strobe_status |= IK_IRQ;
    }
}

/* Strobe NMI. */
_INT_FUNC void strobe_nmi(struct cpu_int_status *cs)
{
    if (!(cs->strobe_status & IK_NMI)) {
	cs->should_check_pending_interrupt++;
	cs->strobe_status |= IK_NMI;
    }
}

/* Remove all strobes. */
_INT_FUNC void remove_strobe(struct cpu_int_status *cs)
{
    if (cs->strobe_status & IK_IRQ)
	cs->should_check_pending_interrupt--;
    if (cs->strobe_status & IK_NMI)
	cs->should_check_pending_interrupt--;
    cs->strobe_status = 0;
}

/* The following functions are used to trigger a RESET or a TRAP.  No function
   to remove these conditions is provided since they are removed as soon as
   they are detected. */

/* RESET. */
_INT_FUNC void trigger_reset(struct cpu_int_status *cs, CLOCK clk)
{
    if (!cs->reset) {
	cs->reset = 1;
	cs->should_check_pending_interrupt++;
	/* Force a pending interrupt check as the next event. */
	cs->next_alarm = -1;
	cs->next_alarm_clk = clk;
    }
}

/* TRAP.  This is a special condition that can be used for debugging.
   `trap_func' will be called with PC as the argument when this condition is
   detected. */
_INT_FUNC void trigger_trap(struct cpu_int_status *cs,
			    void (*trap_func)(ADDRESS), CLOCK clk)
{
    if (!cs->trap) {
	cs->trap = 1;
	cs->trap_func = trap_func;
	cs->should_check_pending_interrupt++;
	/* Force a pending interrupt check as the next event. */
	cs->next_alarm = -1;
	cs->next_alarm_clk = clk;
    }
}

/* Find the next pending alarm. */
_INT_FUNC void find_next_alarm(struct cpu_int_status *cs)
{
    CLOCK next_alarm_clk = ~((CLOCK)0);
    int next_alarm = 0;
    int i;
    
    for (i = 0; i < cs->num_ints; i++)
	if (cs->alarm_clk[i] < next_alarm_clk) {
	    next_alarm = i;
	    next_alarm_clk = cs->alarm_clk[i];
	}

    cs->next_alarm = next_alarm;
    cs->next_alarm_clk = next_alarm_clk;
}

/* Return the clock tick for the next alarm. */
_INT_FUNC CLOCK next_alarm_clk(struct cpu_int_status *cs)
{
    return cs->next_alarm_clk;
}

/* Schedule one cpu alarm. */
_INT_FUNC void set_alarm_clk(struct cpu_int_status *cs, int alarm,
			    CLOCK tick)
{
    if (tick != 0 && tick < cs->next_alarm_clk) {
        cs->next_alarm_clk = cs->alarm_clk[alarm] = tick;
	cs->next_alarm = alarm;
    } else {
	if (tick == 0)
	    tick = ~((CLOCK)0);
	cs->alarm_clk[alarm] = tick;
	if (alarm == cs->next_alarm)
	    find_next_alarm(cs);
    }
}

/* Unschedule one cpu alarm. */
_INT_FUNC void unset_alarm(struct cpu_int_status *cs, int alarm)
{
    cs->alarm_clk[alarm] = ~((CLOCK)0);
    if (alarm == cs->next_alarm)
	find_next_alarm(cs);
}

/* Find whether we have a pending interrupt.  If we have one, return its type.
   Otherwise, return IK_NONE. */
_INT_FUNC int check_pending_interrupt(struct cpu_int_status *cs)
{
    if (cs->should_check_pending_interrupt) {
	if (cs->reset) {
	    /* The `reset' condition is removed as soon as we detect it. */
	    cs->reset = 0;
	    cs->should_check_pending_interrupt--;
	    return IK_RESET;
	} else if (cs->trap) {
	    /* The `trap' condition is removed as soon as we detect it. */
	    cs->trap = 0;
	    cs->should_check_pending_interrupt--;
	    return IK_TRAP;
	} else if ((cs->nnmi && !cs->last_nnmi)
		   || (cs->strobe_status & IK_NMI)) {
	    cs->last_nnmi = cs->nnmi;
	    remove_strobe(cs);
	    return IK_NMI;
	} else if (cs->nirq || (cs->strobe_status & IK_IRQ)) {
	    cs->last_nnmi = cs->nnmi;
	    remove_strobe(cs);
	    return IK_IRQ;
	} else if (!cs->last_nnmi) {
	    /* This should NEVER happen! */
	    fprintf(stderr, "check_pending_interrupt(): "
		    "wrong should_check_pending_interrupt!\n");
	}
    }

    cs->last_nnmi = cs->nnmi;
    return IK_NONE;
}

/* Serve the next pending alarm and update the struct so that we know what
   comes next.  If there is a pending interrupt, return its kind.  Otherwise,
   return IK_NONE. */
_INT_FUNC int serve_next_alarm(struct cpu_int_status *cs, CLOCK clk)
{
    /* If `next_alarm' is -1, we only force a check for pending
       interrupts. (used by TRAP and RESET) */
    if (cs->next_alarm != -1) {
	long offset = clk - cs->alarm_clk[cs->next_alarm];
	cs->alarm_clk[cs->next_alarm] = 0;
	(cs->alarm_handler[cs->next_alarm])(offset);
    } else
	find_next_alarm(cs);
    return check_pending_interrupt(cs);
}

/* This is used to avoid clock counter overflows.  Return 1 if the clock
   counters have been changed (this happens if the clock counter `clk' has
   reached PREVENT_CLK_OVERFLOW_TICK). */
_INT_FUNC int prevent_clk_overflow(struct cpu_int_status *cs, CLOCK *clk)
{
    if (*clk > PREVENT_CLK_OVERFLOW_TICK) {
	int i;
	*clk -= PREVENT_CLK_OVERFLOW_SUB;
	cs->next_alarm_clk -= PREVENT_CLK_OVERFLOW_SUB;
	for (i = 0; i < cs->num_alarms; i++)
	    cs->alarm_clk[i] -= PREVENT_CLK_OVERFLOW_SUB;
	return 1;
    } else
	return 0;
}

#else  /* !INLINE_INTERRUPT_FUNCS && !_MAINCPU_C*/

/* We don't want inline definitions: just provide the prototypes. */

extern void cpu_int_status_init(struct cpu_int_status *cs, int num_ints,
				int num_alarms);
extern void set_irq(struct cpu_int_status *cs, int int_num, int value);
extern void set_nmi(struct cpu_int_status *cs, int int_num, int value);
extern void set_int(struct cpu_int_status *cs, int int_num,
		    enum cpu_int value);
extern void strobe_irq(struct cpu_int_status *cs);
extern void strobe_nmi(struct cpu_int_status *cs);
extern void remove_strobe(struct cpu_int_status *cs);
extern void trigger_reset(struct cpu_int_status *cs, CLOCK clk);
extern void trigger_trap(struct cpu_int_status *cs,
			 void (*trap_func)(ADDRESS), CLOCK clk);
extern void find_next_alarm(struct cpu_int_status *cs);
extern CLOCK next_alarm_clk(struct cpu_int_status *cs);
extern void set_alarm_clk(struct cpu_int_status *cs, int alarm, CLOCK tick);
extern void unset_alarm(struct cpu_int_status *cs, int alarm);
extern int check_pending_interrupt(struct cpu_int_status *cs);
extern int serve_next_alarm(struct cpu_int_status *cs, CLOCK clk);
extern int prevent_clk_overflow(struct cpu_int_status *cs, CLOCK *clk);

#endif /* !INLINE_INTERRUPT_FUNCS && !_MAINCPU_C*/

/* -------------------------------------------------------------------------- */

extern struct cpu_int_status maincpu_int_status;
extern CLOCK clk;

extern struct cpu_int_status true1541_int_status;
extern CLOCK true1541_clk;

/* For convenience... */

#define maincpu_set_alarm_clk(alarm, tick) \
    set_alarm_clk(&maincpu_int_status, (alarm), (tick))
#define maincpu_set_alarm(alarm, ticks_from_now) \
    maincpu_set_alarm_clk((alarm), clk + (ticks_from_now))
#define maincpu_unset_alarm(alarm) \
    unset_alarm(&maincpu_int_status, (alarm))
#define maincpu_set_irq(int_num, value)	\
    set_irq(&maincpu_int_status, (int_num), (value))
#define maincpu_set_nmi(int_num, value) \
    set_nmi(&maincpu_int_status, (int_num), (value))
#define maincpu_set_int(int_num, value) \
    set_int(&maincpu_int_status, (int_num), (value))
#define maincpu_trigger_reset() \
    trigger_reset(&maincpu_int_status, clk + 1)
#define maincpu_strobe_nmi() \
    strobe_nmi(&maincpu_int_status)
#define maincpu_trigger_trap(trap_func) \
    trigger_trap(&maincpu_int_status, (trap_func), clk + 1)
#define maincpu_serve_next_alarm() \
    serve_next_alarm(&maincpu_int_status, clk)
#define maincpu_prevent_clk_overflow() \
    prevent_clk_overflow(&maincpu_int_status, &clk)
	
#define true1541_set_alarm_clk(alarm, tick) \
    set_alarm_clk(&true1541_int_status, (alarm), (tick))
#define true1541_set_alarm(alarm, ticks_from_now) \
    true1541_set_alarm_clk((alarm), true1541_clk + (ticks_from_now))
#define true1541_unset_alarm(alarm) \
    unset_alarm(&true1541_int_status, (alarm))
#define true1541_set_irq(int_num, value)	\
    set_irq(&true1541_int_status, (int_num), (value))
#define true1541_set_nmi(int_num, value) \
    set_nmi(&true1541_int_status, (int_num), (value))
#define true1541_set_int(int_num, value) \
    set_int(&true1541_int_status, (int_num), (value))
#define true1541_trigger_reset() \
    trigger_reset(&true1541_int_status, true1541_clk + 1)
#define true1541_strobe_nmi() \
    strobe_nmi(&true1541_int_status)
#define true1541_trigger_trap(trap_func) \
    trigger_trap(&true1541_int_status, (trap_func), true1541_clk + 1)
#define true1541_serve_next_alarm() \
    serve_next_alarm(&true1541_int_status, true1541_clk)

#endif /* !_INTERRUPT_H */
